/*

------------------------------------------------------------------------------

A license is hereby granted to reproduce this software source code and
to create executable versions from this source code for personal,
non-commercial use.  The copyright notice included with the software
must be maintained in all copies produced.

THIS PROGRAM IS PROVIDED "AS IS". THE AUTHOR PROVIDES NO WARRANTIES
WHATSOEVER, EXPRESSED OR IMPLIED, INCLUDING WARRANTIES OF
MERCHANTABILITY, TITLE, OR FITNESS FOR ANY PARTICULAR PURPOSE.  THE
AUTHOR DOES NOT WARRANT THAT USE OF THIS PROGRAM DOES NOT INFRINGE THE
INTELLECTUAL PROPERTY RIGHTS OF ANY THIRD PARTY IN ANY COUNTRY.

Copyright (c) 1994-2006, John Conover, All Rights Reserved.

Comments and/or bug reports should be addressed to:

    john@email.johncon.com (John Conover)

    http://www.johncon.com/ntropix/
    http://www.johncon.com/

------------------------------------------------------------------------------

spgettermios.c is for getting the termios(3) serial port settings. The
default port is /dev/ttyS0 at 38400 baud. The default semaphore lock
file is /var/lock/LCK..ttyS0, which has an fcntl(2) advisory lock
applied, also.

The program is useful for getting the termios(3) data for the
spsettermios(1) program, and should be run after a "cold boot" of the
machine, when the serial port is in a known state-the spsettermios(1)
program can be used to set the serial port to the "cold boot" known
state on demand.

$Revision: 0.0 $
$Date: 2006/06/18 19:36:00 $
$Id: spgettermios.c,v 0.0 2006/06/18 19:36:00 john Exp $
$Log: spgettermios.c,v $
Revision 0.0  2006/06/18 19:36:00  john
Initial revision


*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <termios.h>
#include <signal.h>
#include <sys/stat.h>

#define _POSIX_SOURCE 1 /* POSIX compliant source */

#define BAUDRATE B38400 /* baud rate for serial port communication */
#define SPDEVICE "/dev/ttyS0" /* default serial port device */
#define LOCKFILE "/var/lock/LCK..ttyS0" /* default serial port lock file name */

#define BUFLEN BUFSIZ /* size of record buffers */

static char rcsid[] = "$Id: spgettermios.c,v 0.0 2006/06/18 19:36:00 john Exp $"; /* program version */
static char copyright[] = "Copyright (c) 1994-2006, John Conover, All Rights Reserved"; /* the copyright banner */

#ifdef __STDC__

static const char *help_message[] = /* help message index array */

#else

static char *help_message[] = /* help message index array */

#endif

{
    "\n",
    "Get the serial port termios\n",
    "Usage: spgettermios [-d device] [-l lockfile] [-v]\n",
    "    -d device, (/dev/ttyS0)\n",
    "    -l lockfile, lock file name, (/var/lock/LCK..ttyS0)\n",
    "    -v, print the program's version information\n"
};

#ifdef __STDC__

static const char *error_message[] = /* error message index array */

#else

static char *error_message[] = /* error message index array */

#endif

{
    "No error\n",
    "Error in program argument(s)\n",
    "Error opening device\n",
    "Error getting device attributes\n",
    "Error flushing device attributes\n",
    "Error closing device\n",
    "Error creating lock file\n",
    "Error writing to lock file\n",
    "Error locking lock file\n",
    "Error unlocking lock file\n",
    "Error removing lock file\n",
    "Error installing signal handler\n",
    "Error closing lock file\n"
    "Error child read process died\n"
};

#define NOERROR 0 /* error values, one for each index in the error message array */
#define EARGS -1
#define EOPEN -2
#define EGATTR -3
#define EFLUSH -4
#define ECLOSE -5
#define ELOCKF -6
#define EWLCK -7
#define EFCNL -8
#define EUFCNL -9
#define ELOCKR -10
#define ESIG -11
#define ELOCKC -12
#define ECHLD -13

#ifdef __STDC__

static void print_message (int pretval); /* print any error messages */
static void spclose (int err); /* interrupt handler */

#else

static void print_message (); /* print any error messages */
static void spclose (); /* interrupt handler */

#endif

static char serial_lockname[] = {LOCKFILE}, /* default lock file name */
            *lockname = serial_lockname; /* default lock file name */

static int retval = NOERROR, /* return value, assume no error */
           fd = -1, /* serial port device file descriptor */
           semaphorelock = -1, /* lock file file descriptor */
           remove_lockfile = 0; /* remove lock file, 0 = no, 1 = yes */

static pid_t pid_write = (pid_t) 0; /* pid of parent write process */

static struct flock lockfile; /* lock file structure */

#ifdef __STDC__

int main (int argc, char *argv[])

#else

int main (argc, argv)
int argc;
char *argv[];

#endif

{
    static char pid_str[BUFLEN]; /* pid string */

    char serial_device[] = {SPDEVICE}, /* default serial port device name */
         *device = serial_device; /* default serial port device name */

    int c; /* command line switch */

    struct termios oldtio; /* original device attributes, restored on exit */

    while ((c = getopt (argc, argv, "d:l:v")) != EOF) /* for each command line switch */
    {

        switch (c) /* which switch? */
        {

            case 'd': /* request for device name? */

                device = optarg; /* yes, save the device name */
                break;

            case 'l': /* request for lock file name? */

                lockname = optarg; /* yes, save the lock file name */
                break;

            case 'v':

                (void) printf ("%s\n", rcsid); /* print the version */
                (void) printf ("%s\n", copyright); /* print the copyright */
                optind = argc; /* force argument error */
                retval = EARGS; /* assume not enough arguments */

            default: /* illegal switch? */

                optind = argc; /* force argument error */
                retval = EARGS; /* assume not enough arguments */
                break;
        }

    }

    pid_write = getpid (); /* save the pid of parent write process */

    if (retval == NOERROR) /* enough arguments? */
    {
        lockfile.l_type = F_WRLCK; /* exclusive write lock the entire file */
        lockfile.l_whence = (short int) SEEK_SET;
        lockfile.l_start = (off_t) 0;
        lockfile.l_len = (off_t) 0;

        retval = ELOCKF; /* assume error creating lock file */

        if ((semaphorelock = open (lockname, O_RDWR | O_CREAT | O_EXCL, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH)) >= 0) /* open the lock file exclusively */
        {
            remove_lockfile = 1; /* set the remove lock file, 0 = no, 1 = yes */
            retval = NOERROR; /* assume no error */
        }

        else if ((semaphorelock = open (lockname, O_RDWR, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH)) >= 0) /* open the lock file if it exists */
        {
            retval = NOERROR; /* assume no error */
        }

        if (retval == NOERROR) /* any errors? */
        {
            retval = EFCNL; /* assume error locking lock file */

            if (fcntl (semaphorelock, F_SETLK, &lockfile) == 0) /* exclusive write lock the lock file */
            {
                remove_lockfile = 1; /* set the remove lock file, 0 = no, 1 = yes */
                retval = EWLCK; /* assume error writing to lock file */

                if (ftruncate (semaphorelock, 0) == 0) /* truncate the lock file */
                {
                    (void) sprintf (pid_str, "%10d\n", pid_write); /* the pid string is the PID right justified to 10 characters plus a '\n' EOL */

                    if (write (semaphorelock, pid_str, strlen (pid_str)) >= 0) /* write the serial port dialer string to the serial port */
                    {
                        retval = ESIG; /* assume error installing signal handler */

                        if (signal (SIGINT, spclose) != SIG_ERR) /* install the interrupt handler for SIGINT = ^C */
                        {

                            if (signal (SIGTERM, spclose) != SIG_ERR) /* install the interrupt handler for SIGTERM */
                            {
                                retval = EOPEN; /* assume error opening device */

                                if ((fd = open (device, O_RDWR | O_NOCTTY)) >= 0) /* open the serial port device file descriptor */
                                {
                                    retval = EFLUSH; /* assume error flushing device attributes */

                                    if (tcflush (fd, TCIOFLUSH) == 0) /* flush device input/output */
                                    {
                                        retval = EGATTR; /* assume error getting device attributes */

                                        if (tcgetattr (fd, &oldtio) == 0) /* save original port settings */
                                        {

                                            /*

                                            taken from "man termios"

                                            */

                                            (void) printf ("oldtio.c_cflag = %d\n", oldtio.c_cflag);
                                            (void) printf ("oldtio.c_iflag = %d\n", oldtio.c_iflag);
                                            (void) printf ("oldtio.c_oflag = %d\n", oldtio.c_oflag);
                                            (void) printf ("oldtio.c_lflag = %d\n", oldtio.c_lflag);
                                            (void) printf ("oldtio.c_cc[VINTR] = %d\n", oldtio.c_cc[VINTR]);
                                            (void) printf ("oldtio.c_cc[VQUIT] = %d\n", oldtio.c_cc[VQUIT]);
                                            (void) printf ("oldtio.c_cc[VERASE] = %d\n", oldtio.c_cc[VERASE]);
                                            (void) printf ("oldtio.c_cc[VKILL] = %d\n", oldtio.c_cc[VKILL]);
                                            (void) printf ("oldtio.c_cc[VEOF] = %d\n", oldtio.c_cc[VEOF]);
                                            (void) printf ("oldtio.c_cc[VTIME] = %d\n", oldtio.c_cc[VTIME]);
                                            (void) printf ("oldtio.c_cc[VMIN] = %d\n", oldtio.c_cc[VMIN]);
                                            (void) printf ("oldtio.c_cc[VSWTC] = %d\n", oldtio.c_cc[VSWTC]);
                                            (void) printf ("oldtio.c_cc[VSTART] = %d\n", oldtio.c_cc[VSTART]);
                                            (void) printf ("oldtio.c_cc[VSTOP] = %d\n", oldtio.c_cc[VSTOP]);
                                            (void) printf ("oldtio.c_cc[VSUSP] = %d\n", oldtio.c_cc[VSUSP]);
                                            (void) printf ("oldtio.c_cc[VEOL] = %d\n", oldtio.c_cc[VEOL]);
                                            (void) printf ("oldtio.c_cc[VREPRINT] = %d\n", oldtio.c_cc[VREPRINT]);
                                            (void) printf ("oldtio.c_cc[VDISCARD] = %d\n", oldtio.c_cc[VDISCARD]);
                                            (void) printf ("oldtio.c_cc[VWERASE] = %d\n", oldtio.c_cc[VWERASE]);
                                            (void) printf ("oldtio.c_cc[VLNEXT] = %d\n", oldtio.c_cc[VLNEXT]);
                                            (void) printf ("oldtio.c_cc[VEOL2] = %d\n", oldtio.c_cc[VEOL2]);

                                            retval = EFLUSH; /* assume error flushing device attributes */

                                            if (tcflush (fd, TCIOFLUSH) == 0) /* flush device input/output */
                                            {
                                                retval = NOERROR; /* assume no error */
                                            }

                                        }

                                    }

                                }

                            }

                        }

                    }

                }

            }

        }

    }

    spclose (0); /* shutdown */
    exit (retval); /* exit with the error value, (formality: spclose () does the shutdown, and exit) */


#ifdef LINT

    return (0); /* for lint formalities */

#endif

}

/*

Print any error messages.

static void print_message (int pretval);

I) Data structures:

    A) The help_message array is an array of character stings, one per
       line to be printed for help.

    B) The error_message array is an array of character strings, one
       line per error; the error_message array is implicitly addressed
       by the value integer pretval, which specifies the error to be
       printed.

II) Function execution:

    A) Depending on the value of the integer pretval:

        1) If pretval is zero, print nothing-normal/successful program
           exit.

        2) Else if pretval is unity, print help.

        3) Else pretval is not zero or unity, it is an error code, print
           the corresponding error message.

Returns nothing.

*/

#ifdef __STDC__

static void print_message (int pretval)

#else

static void print_message (pretval)
int pretval;

#endif

{
    size_t help_ctr; /* help_message line counter */

    switch (pretval) /* which return value */
    {

        case 0: /* program ended without errors, print nothing */

            break;

        case 1: /* program ended with a request for help, or argument error, print help */

            for (help_ctr = 0; help_ctr < (sizeof (help_message) / sizeof (char *)); help_ctr ++) /* for each line of help */
            {
                (void) printf ("%s", help_message[help_ctr]); /* print the line */
            }

            break;

        default: /* an error that was not a request for help, print the error */

            (void) fprintf (stderr, "%s", error_message[pretval]);
            break;

    }

}

/*

Interrupt handler and shutdown.

static void spclose (int err);

I) Prerequisites:

    A) All file opening and serial port initialization is done by the
       this process, and this process should close all files, and
       reset the serial port.

    B) Called with SIGINT = ^C, or, SIGTERM interrupts, or, with a
       negative argument representing an internal error, or, an
       argument of zero for no errors or interrupts.

    C) There are "edge" issues, (i.e., a flag is tested to see if
       something needs shutting down, and the something is shut down,
       but the program is interrupted-perhaps by SIGINT = ^C
       bashing-before the flag gets reset.) There are no error message
       printed in the shut down section of this routine-everything is
       shut down, and at the end of this routine, the flags are
       queried for a successful shutdown, and shutdown errors, if any,
       printed. The idea being that if an "edge" issue occurred, the
       shut down would occur twice, (presumably giving an error the
       second time, which would not be printed, but as long as the
       flag is reset, no error would be printed.)

II) Function execution:

    A) Upon entry:

        1) Shut down all devices opened by the parent write process.

        2) Print any pending errors that occurred before the program
           entered this routine.

        3) Print any errors after the shutdown is complete.

        4) Restore the signal handlers for SIGINT = ^C and SIGTERM.

There is no return from this routine. The routine exits with a value
of zero if no errors or interrupts occurred, or, the value of the
interrupt, (a positive number,) if an interrupt occurred, or, with a
negative error number if an an error occurred.

*/

#ifdef __STDC__

static void spclose (int err)

#else

static void spclose (err)
int err;

#endif

{
    int exit_value,
        local_retval = NOERROR; /* assume no error shutting things down */

    if (fd > 0) /* serial port device file descriptor successfully opened? */
    {
        (void) tcflush (fd, TCIOFLUSH); /* flush device input/output */

        if (close (fd) == 0) /* yes, close the device */
        {
            fd = -1; /* reset the serial port device file descriptor */
        }

    }

    if (semaphorelock >= 0) /* lock file open? */
    {
        lockfile.l_type = F_UNLCK; /* unlock the file */
        (void) fcntl (semaphorelock, F_SETLKW, &lockfile); /* unlock the lock file */

        if (close (semaphorelock) == 0) /* close the lock file */
        {
            semaphorelock = -1; /* reset the lock file file descriptor */

            if (remove_lockfile == 1) /* remove lock file, 0 = no, 1 = yes, set? */
            {

                if (unlink (lockname) == 0) /* remove the lock file */
                {
                    remove_lockfile = 0; /* remove lock file, 0 = no, 1 = yes */
                }

            }

        }

    }

    print_message (-retval); /* print any error messages from the program, (there can only be zero or one) */

    /* print any errors shutting things down */

    if (semaphorelock != -1) /* lock file open? */
    {print_message (-ELOCKR); local_retval = ELOCKR;} /* print the error message, set the local error shutting things down */

    /* probably don't care about these-exit () will handle them */

    if (fd != -1) /* serial port device file descriptor still open? */
    {print_message (-ECLOSE); local_retval = ECLOSE;} /* print the error message, set the local error shutting things down */

    /* probably don't care about these-exit () will handle them */

    (void) signal (SIGTERM, SIG_DFL); /* restore default interrupt for SIGTERM */
    (void) signal (SIGINT, SIG_DFL); /* restore default for SIGINT = ^C interrupt */

    exit_value = err; /* assume routine called externally, i.e., SIGINT == ^C or SIGTERM */

    if (exit_value == 0) /* no? */
    {
        exit_value = retval; /* assume error in main body of program, before this routine */

        if (retval == 0) /* no? */
        {
            exit_value = local_retval; /* assume error shutting things down */
        }

    }

    exit (exit_value); /* exit with the value of the interrupt, the value of an error in the main body of this program, the value of an error in this routine, or zero for no error */
}
